/*
 * buffersync.h - Buffer sync.
 */
#ifndef SIRI_BUFFERSYNC_H_
#define SIRI_BUFFERSYNC_H_

#include <siri/siri.h>

void siri_buffersync_init(siri_t * siri);
void siri_buffersync_stop(siri_t * siri);

#endif  /* SIRI_HEARTBEAT_H_ */
