/*
 * variance.h - Calculate variance for points.
 */
#ifndef SIRIDB_VARIANCE_H_
#define SIRIDB_VARIANCE_H_

#include <siri/db/points.h>

double siridb_variance(siridb_points_t * points);

#endif  /* SIRIDB_VARIANCE_H_ */
