/*
 * bserver.h - Listen to back-end SiriDB Server.
 */
#ifndef SIRINET_BSERVER_H_
#define SIRINET_BSERVER_H_

#include <siri/siri.h>

int sirinet_bserver_init(siri_t * siri);

#endif  /* SIRINET_BSERVER_H_ */
